import api
import ui
import scriptHandler
import IAccessibleHandler
import speech
import textInfos
from textInfos import offsets
from NVDAObjects import behaviors
import re
from NVDAObjects.IAccessible import IAccessible,MenuItem
from ctypes import *
import winUser
import keyboardHandler
import sayAllHandler

class VsTextPos(IAccessible):
	"""This class encapsulates an IAccessible representing character object somewhere in the VS code editor."""
	def __init__(self,**kwargs):
		super(VsTextPos,self).__init__(**kwargs)
		self._lineNum=self._colNum=None

	def _numFromName(self,name):
		"""Returns a number from object's accName like 'Line 1000' or 'Character 7'"""
		m=re.match(".*?(\d+)",name)
		if not m: return -1
		return int(m.group(1))

	def _get_lineNum(self):
		if not self._lineNum:
			self._lineNum=self._numFromName(self.parent.name)
		return self._lineNum

	def _get_colNum(self):
		if not self._colNum:
			self._colNum=self._numFromName(self.name)
		return self._colNum

	def __cmp__(self,other):
		if not isinstance(other,VsTextPos): raise NotImplementedError(repr(other))
		if self.lineNum<other.lineNum: return -1
		if self.lineNum>other.lineNum: return 1
		if self.colNum<other.colNum: return -1
		if self.colNum>other.colNum: return 1
		return 0

	def __eq__(self,other):
		if not isinstance(other,VsTextPos) and not isinstance(other,basestring): raise NotImplementedError(repr(other))
		if isinstance(other,basestring): return False
		return self.lineNum==other.lineNum and self.colNum==other.colNum

	def _get_value(self):
		"""Override value because the default implementation returns None if value is only whitespace. This is not what we want because for example the caret may land on whitespace."""
		return self.IAccessibleObject.accValue(self.IAccessibleChildID)

	def _get_line(self):
		"""Returns a line this character belongs to as a text."""
		return self.IAccessibleObject.accValue(0)

	def __repr__(self):
		return"VsTextPos(%d, %d)"%(self.lineNum,self.colNum)

class VsTextRange(object):
	"""Represents two absolute positions in VsTextInfo."""
	def __init__(self,startPos,endPos):
		self.startPos=startPos
		self.endPos=endPos

class VsTextInfo(textInfos.TextInfo):
	def __init__(self,obj,position):
		super(VsTextInfo,self).__init__(obj,position)
		self._startPos=self._endPos=None
		if isinstance(position,VsTextRange):
			self._startPos=position.startPos
			self._endPos=position.endPos
		elif isinstance(position,basestring):
			if position==textInfos.POSITION_FIRST: self._startPos=self._endPos=self.obj.charAt(1,1)
			elif position==textInfos.POSITION_LAST: self._startPos=self._endPos=self.obj.charAt(obj.childCount,1)
			elif position==textInfos.POSITION_CARET: self._startPos=self._endPos=self.obj.caretPos
			elif position==textInfos.POSITION_SELECTION:
				startLine,startCol,endLine,endCol=self.obj.selectionInfo
				self._startPos=self.obj.charAt(startLine,startCol)
				self._endPos=self.obj.charAt(endLine,endCol)
			elif position==textInfos.POSITION_ALL:
				self._startPos=self._endPos="all" #special case, we simply retrieve whole text from obj.value
		else: raise NotImplementedError("position: %s not supported"%position)

	def __eq__(self,other):
		if isinstance(other,VsTextInfo) and self._startPos==other._startPos and self._endPos==other._endPos: return True
		return False

	def _get_isCollapsed(self):
		return self._startPos==self._endPos

	def collapse(self,end=False):
		if self._startPos=="all":
			if not end: self._startPos=self._endPos=self.obj.charAt(1,1)
			else: self._startPos=self._endPos=self.obj.charAt(self.obj.childCount,1)
		else:
			if not end: self._endPos=self._startPos
			else: self._startPos=self._endPos

	def _getWordPositions(self,vsTextPos):
		line=vsTextPos.line
		lineNum=vsTextPos.lineNum
		colNum=vsTextPos.colNum-1
		startOfWord=offsets.findStartOfWord(line,colNum)
		endOfWord=offsets.findEndOfWord(line,colNum)
		return (self.obj.charAt(lineNum,startOfWord+1),self.obj.charAt(lineNum,endOfWord+1))

	def expand(self,unit):
		if unit==textInfos.UNIT_CHARACTER: self._endPos=self._movePos(self._startPos,unit,1)
		elif unit==textInfos.UNIT_WORD: self._startPos,self._endPos=self._getWordPositions(self._startPos)
		elif unit==textInfos.UNIT_LINE:
			self._startPos=self.obj.charAt(self._startPos.lineNum,1)
			self._endPos=self.obj.charAt(self._startPos.lineNum,self._startPos.parent.childCount)
	def compareEndPoints(self,other,which):
		if which=="startToStart": return cmp(self._startPos,other._startPos)
		elif which=="startToEnd": return cmp(self._startPos,other._endPos)
		elif which=="endToStart": return cmp(self._endPos,other._startPos)
		elif which=="endToEnd": return cmp(self._endPos,other._endPos)
		else: raise ValueError("bad argument - which: %s"%which)

	def setEndPoint(self,other,which):
		if which=="startToStart": self._startPos=other._startPos
		elif which=="startToEnd": self._startPos=other._endPos
		elif which=="endToStart": self._endPos=other._startPos
		elif which=="endToEnd": self._endPos=other._endPos
		else: raise ValueError("bad argument - which: %s"%which)

	def _get_text(self):
		if self.isCollapsed: return ""
		if self._startPos=="all": return self.obj.value
		if self._startPos.lineNum==self._endPos.lineNum:
			line=self._startPos.line
			return line[self._startPos.colNum-1:self._endPos.colNum-1]
		#Retrieve a range of lines
		lines=[]
		startLine,startCol=self._startPos.lineNum,self._startPos.colNum
		endLine,endCol=self._endPos.lineNum,self._endPos.colNum
		#Retrieve rest of first line
		lines.append(self._startPos.line[startCol-1:])
		i=startLine+1
		while i<endLine:
			lines.append(self.obj.charAt(i,1).line)
			i+=1
		#Retrieve start of last line
		lines.append(self.obj.charAt(i,1).line[:endCol-1])
		return "\r\n".join(lines)

	def _get_bookmark(self):
		return VsTextRange(self._startPos,self._endPos)

	def copy(self):
		return VsTextInfo(self.obj,self.bookmark)

	def _movePos(self,vsTextPos,unit,step):
		"""Moves a VsTextPos instance either backward or forward one step by unit, depending on step.
		If step is 1, move forward, if step is -1, move backward."""
		lineNum=vsTextPos.lineNum
		colNum=vsTextPos.colNum
		lineLength=vsTextPos.parent.childCount
		lineCount=self.obj.childCount
		if unit==textInfos.UNIT_CHARACTER:
			colNum+=step
			if colNum==0:
				lineNum-=1
				colNum=1
			elif colNum>lineLength:
				lineNum+=1
				colNum=1
			if lineNum==0 or lineNum>lineCount: return None
			return self.obj.charAt(lineNum,colNum)
		elif unit==textInfos.UNIT_WORD:
			startOfWord,endOfWord=self._getWordPositions(vsTextPos)
			if step==-1:
				pos=self._movePos(startOfWord,textInfos.UNIT_CHARACTER,-1)
				if pos is None: pos=self.obj.charAt(1,1)
				sw,ew=self._getWordPositions(pos)
				return sw
			else:
				pos=self._movePos(endOfWord,textInfos.UNIT_CHARACTER,1)
				if pos is None: pos=self.obj.charAt(self.obj.childCount,1)
				sw,ew=self._getWordPositions(pos)
				return sw
		elif unit==textInfos.UNIT_LINE or unit==textInfos.UNIT_READINGCHUNK:
			lineNum+=step
			if lineNum==0 or lineNum>lineCount: return None
			return self.obj.charAt(lineNum,1)

	def move(self,unit,direction,endPoint=None):
		if direction==0: return 0
		if endPoint=="start": pos=self._startPos
		elif endPoint=="end": pos=self._endPos
		else:
			self.collapse()
			pos=self._startPos
		count=0
		step=1 if direction>0 else -1
		while count!=direction:
			pos=self._movePos(pos,unit,step)
			if pos is None: return count
			if endPoint=="start": self._startPos=pos
			elif endPoint=="end": self._endPos=pos
			else: self._startPos=self._endPos=pos
			count+=step
		return count

	def updateCaret(self):
		"""This is here only to support updating caret during sayAll. It is probably not
		possible to move caret to any position without some insane algorithm requiring sending keystrokes.
		Here we send the downArrow key when sayAll.isRunning() returns True. Else raise NotImplementedError."""
		if sayAllHandler.isRunning(): keyboardHandler.KeyboardInputGesture.fromName("downArrow").send()
		else: raise NotImplementedError
		

class VsTextEditPane(behaviors.EditableTextWithoutAutoSelectDetection):
	TextInfo=VsTextInfo

	def charAt(self,lineNum,colNum):
		"""Returns a VsTextPos object representing character at position lineNum/colNum"""
		lineCount=self.childCount
		if lineNum>lineCount: lineNum=lineCount
		try:
			line=self.IAccessibleObject.accChild(lineNum+1)
		except:
			line=None
		if not line: return None
		colCount=line.accChildCount
		if colCount==0: colNum=1
		if colNum>colCount: colNum=colCount
		return VsTextPos(IAccessibleObject=line,IAccessibleChildID=colNum)

	def _get_caretPos(self):
		caretRect=winUser.getGUIThreadInfo(self.windowThreadID).rcCaret
		p=winUser.POINT()
		p.x=caretRect.left
		p.y=caretRect.top
		winUser.user32.ClientToScreen(self.windowHandle,byref(p))
		ia,childID=IAccessibleHandler.accessibleObjectFromPoint(p.x,p.y)
		global a,i
		a,i=ia,childID
		return VsTextPos(IAccessibleObject=ia,IAccessibleChildID=childID)

	def _get_selectionInfo(self):
		"""Returns information about selected text by parsing description of IAccessible object representing selection like 'stream from line 0, char 0 to line 1, char 0'."""
		nums=".*?(\d+)"*4
		m=re.match(nums,self.firstChild.description)
		if not m: return []
		res=[]
		for i in range(4):
			res.append(int(m.group(i+1))+1)
		return res

class MsoCommandBarPopup(MenuItem):
	def _get_name(self):
		return self.displayText


